package com.ca.smdeletesession;


import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Properties;
import java.util.TimeZone;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.GZIPOutputStream;

import javax.naming.Context;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;
import javax.naming.directory.SearchControls;
import javax.naming.ldap.InitialLdapContext;
import javax.naming.ldap.LdapContext;

import com.sf.smutils.util.RuntimeProperties;
import com.unboundid.asn1.ASN1OctetString;
import com.unboundid.ldap.sdk.Attribute;
import com.unboundid.ldap.sdk.BindResult;
import com.unboundid.ldap.sdk.DeleteRequest;
import com.unboundid.ldap.sdk.DereferencePolicy;
import com.unboundid.ldap.sdk.Filter;
import com.unboundid.ldap.sdk.LDAPConnection;
import com.unboundid.ldap.sdk.LDAPConnectionOptions;
import com.unboundid.ldap.sdk.LDAPException;
import com.unboundid.ldap.sdk.LDAPResult;
import com.unboundid.ldap.sdk.ResultCode;
import com.unboundid.ldap.sdk.SearchRequest;
import com.unboundid.ldap.sdk.SearchResult;
import com.unboundid.ldap.sdk.SearchResultEntry;
import com.unboundid.ldap.sdk.SearchScope;
import com.unboundid.ldap.sdk.controls.SimplePagedResultsControl;
import com.unboundid.ldap.sdk.controls.SubtreeDeleteRequestControl;
import com.unboundid.util.ssl.SSLUtil;
import com.unboundid.util.ssl.TrustAllTrustManager;


public class SessionStore_LdapDelete {

	
	//public class Main {
	public static void main(String[] args) {
		
		try {
			
		    SessionStore_LdapDelete main = new SessionStore_LdapDelete();
			main.runMain();
			
		}
		catch (Exception e) {
			e.printStackTrace();
		}
			
	}

    boolean debugTrace = false;
    boolean debugTraceAll = false;
    boolean ldapTestMode = false;
    boolean ldapUsePageControl = true; 
	
    
    LDAPConnection ldapConnection = null;
    
    void runMain() {

        getPolicyStoreConnectProperties();
        
        debugTrace = connectProperties.getPropertyAsBoolean("debugTrace", true);
        debugTraceAll = connectProperties.getPropertyAsBoolean("debugTraceAll", false);
        ldapTestMode = connectProperties.getPropertyAsBoolean("ldapTestMode", false);
        ldapUsePageControl = connectProperties.getPropertyAsBoolean("ldapUsePageControl", true);
         
        
        //if (isStillRunning()) {
        //    if (debugTrace) {
        //        System.out.println("Previous run is still going or completed recently - exiting");
        //    }
        //    return;
        //}
        touchFileTimeStamp();               
        
        connect();

        // If we are using page control - it will loop through all the entries found 
        if (ldapUsePageControl) {
            runSearchAndDelete();
        }
        else { 
            do { 
                runSearchAndDelete();
            }
            while (totalDeletedCount > 0); 
        }
        
        ldapConnection.close();
        
    }


    File keepAliveFile = null;
    
    File getKeepAliveFile() {
        
        if (keepAliveFile == null) {
            String keepAliveFileName  = connectProperties.getProperty("keepAliveFileName", ".keepAliveFile");
        
            keepAliveFile = new File(keepAliveFileName);
        }
        return keepAliveFile;
    }
    
    
    boolean isStillRunning() {

        File file = getKeepAliveFile();
        
        if (!file.exists()) {
             return false;
        }
        
        long lastModified = file.lastModified();
        long appendTime = connectProperties.getPropertyAsLong("CheckTimeGapMin", 5 );
        
        appendTime = appendTime * 60 * 1000; 

        lastModified += appendTime;
        
        if (lastModified <= System.currentTimeMillis()) {
            return false;
        }
        
        return true;
    }
    
    /**
     * Update timestamp on the file.
     */
    
    void touchFileTimeStamp() {
        File file = getKeepAliveFile();

        try { 
            FileWriter fw = new FileWriter(file);
            Date date = new Date(System.currentTimeMillis());
            
            fw.append(date.toString());
            fw.close();
        }
        catch (IOException e) {}
        
    }
        
    void connect() {

        String providerUrl = connectProperties.getProperty("ldapProviderUrl");
        String userDn = connectProperties.getProperty("ldapUserDn", null);
        String password = connectProperties.getProperty("ldapUserPassword", null);
        boolean followReferrals = connectProperties.getPropertyAsBoolean("followReferrals", false);

        try { 

            Pattern urlPattern = Pattern.compile("(\\w+)://(\\S+):(\\w+)");
            
            Matcher matcher = urlPattern.matcher(providerUrl);
            
            if (!matcher.find()) {
                throw new RuntimeException("Unable to parse ldap[s]://<hostaddress>:<port> from :" + providerUrl);
            }
            
            String protocol = matcher.group(1);
            String address = matcher.group(2);
            int port = 389;
            
            try { 
                String strPort = matcher.group(3);
                port = Integer.parseInt(strPort);
                
            }
            catch (NumberFormatException e) {
                throw new RuntimeException("Unable to parse port from " + providerUrl, e);
            }
            
            boolean isSSL = protocol.toLowerCase().equals("ldaps");
            
            
            if (isSSL) {
                 SSLUtil sslUtil = new SSLUtil(new TrustAllTrustManager());
                ldapConnection =    new LDAPConnection(sslUtil.createSSLSocketFactory());
            }
            else {
                ldapConnection = new LDAPConnection();
            }

            ldapConnection.connect(address, port);
            BindResult bindResult = ldapConnection.bind(userDn, password);

            LDAPConnectionOptions options = new LDAPConnectionOptions();
            
            options.setFollowReferrals(followReferrals);
            
            ldapConnection.setConnectionOptions(options);
            
            String controlOID = SimplePagedResultsControl.PAGED_RESULTS_OID;
            //  if(!SupportedFeature.isControlSupported(ldapConnection,controlOID)) {
            //    return ResultCode.UNWILLING_TO_PERFORM;
            //  }

        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }
	
	RuntimeProperties connectProperties = null;
	
    public RuntimeProperties getPolicyStoreConnectProperties() {
        if (connectProperties == null) {
            
            Properties properties = new Properties();

            File file = new File("policyStoreConnect.properties");
            if (file.exists()) {
                try {
                    properties.load(new FileInputStream(file));
                } catch (FileNotFoundException e) {
                } catch (IOException e) {
                }
            }
            connectProperties = new RuntimeProperties();
            connectProperties.setProperties(properties);
        }
        
        return connectProperties;
    }

	int timeLimit = 30;
	
	long transTime = 0;
	
	long blockTransTime = 0;
	long totalTransTime = 0;

    int totalDeletedCount = 0; 
    int blockDeletedCount = 0; 
	
	
    SearchRequest buildSearchRequest(String baseDN, String search) throws LDAPException {
		SearchScope scope = SearchScope.SUB; //SearchScope.SUBORDINATE_SUBTREE;
	      Filter filter = Filter.create(search);
	      
	      SearchRequest searchRequest;
          searchRequest = new SearchRequest(baseDN,scope,filter,SearchRequest.NO_ATTRIBUTES); 
          //String[] requestedAttributes = new String[0];
	      //searchRequest = new SearchRequest(baseDN,scope,filter,requestedAttributes);
	      int sizeLimit = 0; 
  	      timeLimit = connectProperties.getPropertyAsInt("timeLimit", 30); 		  
	      searchRequest.setSizeLimit(sizeLimit);
	      searchRequest.setTimeLimitSeconds(timeLimit);

	      /*  Should not be needed - we identified the problem.
	       *  
	      // Determine deref policy 
          String deRefPolicyStr = connectProperties.getProperty("deRefPolicy", "Never");
          deRefPolicyStr = deRefPolicyStr.toLowerCase();
          
          DereferencePolicy deRefPolicy =  DereferencePolicy.NEVER;
          if (deRefPolicyStr.startsWith("a")) {
              deRefPolicy = DereferencePolicy.ALWAYS;
          }
          else if (deRefPolicyStr.startsWith("f")) {
              deRefPolicy = DereferencePolicy.FINDING;
          }
          else if (deRefPolicyStr.startsWith("s")) {          
              deRefPolicy = DereferencePolicy.SEARCHING;
          }          
	      
	      searchRequest.setDerefPolicy(deRefPolicy);
	      */
	      
	      return searchRequest;
    }

	
    ArrayList<String> deleteList = new ArrayList<String> ();
    
    void runSearchAndDelete() { 
    	try { 
    	    
    	    String ldapDeleteRoot = connectProperties.getProperty("ldapDeleteRoot");
            String ldapDeleteSearch = connectProperties.getProperty("ldapDeleteSearch");

    		int pageSize = connectProperties.getPropertyAsInt("pageSize", 1000);

    		// If we find %TIME then substitute the current LDAP GMT time 
    		if (ldapDeleteSearch.indexOf("%TIME") != -1) {
                ldapDeleteSearch = ldapDeleteSearch.replace("%TIME", getCurrentTimeAsLdap());
    		}
    		
    		SearchRequest searchRequest = buildSearchRequest(ldapDeleteRoot, ldapDeleteSearch);

            if (debugTrace) { 
                System.out.println("============== Starting Search  ====================");
                System.out.println("");
			    System.out.println("Delete   Root: " + ldapDeleteRoot );
				System.out.println("Delete Search: " + ldapDeleteSearch);
                System.out.println("");
                System.out.println("pageSize: " + pageSize + " timeLimit:" + timeLimit + " sec");
            }
    		
    		ASN1OctetString cookie = null;
    		int total = 0;

			totalDeletedCount = 0;
			totalTransTime = 0;
			
    		do {
    		    
    		    touchFileTimeStamp();    		    
    			/*
    			 * Set the simple paged results control (if the cookie is null
    			 * this indicates the first time through the loop).
    			 */
    			final SimplePagedResultsControl simplePagedResultsRequestControl = new SimplePagedResultsControl(pageSize,cookie);
    			
    			if (ldapUsePageControl) { // Optional to use page control
    			    searchRequest.setControls(simplePagedResultsRequestControl);
    			}

    			/*
    			 * Issue the search request:
    			 */
    			SearchResult searchResult = ldapConnection.search(searchRequest);

    		    deleteList = new ArrayList<String> ();

    		    blockTransTime = 0;    		    
    		    blockDeletedCount = 0;

    			if (debugTraceAll) {
                    String msg = String.format("searchRequest transmitted, " + "entries returned: %d", Integer.valueOf(searchResult.getEntryCount()));
    			    if (ldapUsePageControl) {
    			        msg += ", page size : " +  pageSize; 
    			    }
   			        System.out.println("");
                    System.out.println(msg);
                }

    			
    			deleteEntries(searchResult);
                
    			total += searchResult.getEntryCount();

    			if (debugTraceAll) { 
   			        System.out.println("");
   			        
   			        long avBlockTransTime = (blockDeletedCount == 0) ? 0 : (blockTransTime / blockDeletedCount); 
  			        System.out.println("Entries read " + total + " Entries deleted " + totalDeletedCount + " this page" + blockDeletedCount + " av (" + avBlockTransTime + "ms)");
                }
				
    			// Get the cookie from the paged results control.
    			cookie = null;
    			if (ldapUsePageControl) {
    			    final SimplePagedResultsControl c = SimplePagedResultsControl.get(searchResult);
                    if(c != null) {
                        cookie = c.getCookie();
                    }
    			}

    			// Without page control cookie will always be null 
    		} while((cookie != null) && (cookie.getValueLength() > 0));

			if (debugTrace) {

                long avTotalTransTime = (totalDeletedCount == 0) ? 0 : (totalTransTime / totalDeletedCount); 
			    
			    System.out.println("");
                System.out.println("Total Entries read " + total + " Entries deleted " + totalDeletedCount + " av (" + avTotalTransTime + "ms)" );
			    System.out.println("============== Ending  Search  ====================");
			}
			
    	}
    	catch (Exception e) {
    		throw new RuntimeException(e);
    	}
    }

    SimpleDateFormat sdf = null; 

    SimpleDateFormat getLDAPDateFormat() {
        if (sdf == null) {
            sdf = new SimpleDateFormat("yyyyMMddHHmmss");
            sdf.setTimeZone(TimeZone.getTimeZone("GMT"));
        }
        return sdf;
        
    }
    public String getCurrentTimeAsLdap() {
        Date date = new Date();
        
        String dateStr = getLDAPDateFormat().format(date);
        dateStr +="Z";
        
        return dateStr;
    }
    
    public Date parseLdapDate(String ldapDate){

        try {
            return getLDAPDateFormat().parse(ldapDate);
        } catch (ParseException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        return null;
    }
    
    void deleteEntries(SearchResult searchResult ) {
        
        for (SearchResultEntry searchEntry : searchResult.getSearchEntries()) {
    
            String dn = searchEntry.getDN();
            ldapDelete(dn);            
        }
    }
    
    
   	
	void deleteEntries() { 
	    for (String dn : deleteList) {
	        
	        long startTime = System.currentTimeMillis();
	        ldapDelete(dn);
	        	        
	        long timeTaken = System.currentTimeMillis() - startTime;
	        if (debugTrace) {
	            System.out.println("deleted " + dn + " ("  + timeTaken + "ms)"); 
	        }
	    }
	}


    
    void ldapDelete(String dn) {

        // Test mode we only report the entry not delete it.
        if (ldapTestMode) {  
            System.out.println("not deleted entry (test mode) :" + dn);
            totalDeletedCount++;
            return;
        }
        
        LDAPResult deleteResult = null;
        try  {
            
            long startTime = System.currentTimeMillis();
            DeleteRequest deleteRequest = new DeleteRequest(dn);
            deleteRequest.addControl(new SubtreeDeleteRequestControl());
            deleteResult = ldapConnection.delete(deleteRequest);

			
			transTime = System.currentTimeMillis() - startTime;
			blockTransTime += transTime;
			totalTransTime += transTime;
            blockDeletedCount++;
            totalDeletedCount++;
			
			if (debugTraceAll) { 
			    System.out.println("deleted entry:" + dn + "(" + transTime + "ms");
			}
        }
        catch (LDAPException le)  {
            // The delete operation failed.
            deleteResult = le.toLDAPResult();
            ResultCode resultCode = le.getResultCode();
            String errorMessageFromServer = le.getDiagnosticMessage();
            
            System.out.println("Error : " + resultCode + ": " +  errorMessageFromServer);
        } 
        
    }    

	private static SearchControls getSimpleSearchControls() {
	    SearchControls searchControls = new SearchControls();
	    searchControls.setSearchScope(SearchControls.SUBTREE_SCOPE);
	    searchControls.setCountLimit(0);
	    searchControls.setTimeLimit(30000);
	    //String[] attrIDs = {"objectGUID"};
	    //searchControls.setReturningAttributes(attrIDs);
	    return searchControls;
	}		
	
    void XrunMain() {

        getPolicyStoreConnectProperties();        
        debugTrace = connectProperties.getPropertyAsBoolean("debugTrace", true);
        
        connect();

        while (true) {
            XrunSearch();
        
            if (deleteList.size() == 0) {
                break;
            }
            deleteEntries();
        }
        
        ldapConnection.close();
        
    }
    
    void XrunSearch() { 
        try { 
            
            String ldapDeleteRoot = connectProperties.getProperty("ldapDeleteRoot");
            String ldapDeleteSearch = connectProperties.getProperty("ldapDeleteSearch");
            
            
            //NamingEnumeration<?> namingEnum = ctx.search(getPolicySrv4Path(), "(objectclass=*)", getSMOBJSearchControls());

            SearchRequest searchRequest = buildSearchRequest(ldapDeleteRoot, ldapDeleteSearch);


            if (debugTrace) { 
                System.out.println("============== Starting Search  ====================");
            }
            
            ASN1OctetString cookie = null;
            int total = 0;
            int pageSize = 1000;

            do {
                /*
                 * Set the simple paged results control (if the cookie is null
                 * this indicates the first time through the loop).
                 */
                final SimplePagedResultsControl simplePagedResultsRequestControl = new SimplePagedResultsControl(pageSize,cookie);
                searchRequest.setControls(simplePagedResultsRequestControl);

                /*
                 * Issue the search request:
                 */
                SearchResult searchResult = ldapConnection.search(searchRequest);

                deleteList = new ArrayList<String> ();
                
                String msg = String.format("searchRequest transmitted, pageSize: %d, " + "entries returned: %d",Integer.valueOf(pageSize), Integer.valueOf(searchResult.getEntryCount()));

                if (debugTrace) { 
                    System.out.println(msg);
                }

                for (SearchResultEntry searchEntry : searchResult.getSearchEntries()) {
                    String dn = searchEntry.getDN();
                    
                    deleteList.add(dn);
                    
                }
                deleteEntries();
                
                
                total += searchResult.getEntryCount();

                // Get the cookie from the paged results control.
                cookie = null;
                final SimplePagedResultsControl c = SimplePagedResultsControl.get(searchResult);
                if(c != null) {
                    cookie = c.getCookie();
                }
            } while((cookie != null) && (cookie.getValueLength() > 0));
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }
    
	
}

