package com.sf.smutils.util;

import java.security.*;

import javax.crypto.*;
import javax.crypto.spec.*;


import java.io.*;

public class DataEncryptor {
    SecretKeySpec secretKeySpec;
    Cipher cipher;

    public DataEncryptor()     {
        try { 
            byte[] raw = generateKey();
            secretKeySpec = new SecretKeySpec(raw, "AES");
            cipher = Cipher.getInstance("AES/ECB/PKCS5Padding");
            //Cipher cipher = Cipher.getInstance("AES/ECB/PKCS7Padding");
            //Cipher cipher = Cipher.getInstance("AES/ECB/PKCS7Padding", "BC");
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    } 

    String keyFileName = ".client_encrypt_key.txt";

    byte[] generateKey() throws Exception
    {

        File appDir = new File(".");

        File file = new File(appDir, keyFileName);

        if (file.exists()) {
            return Base64.decodeFromFile(file.getAbsolutePath());
        }

        // So does not exist so we create it.

        KeyGenerator keyGen = KeyGenerator.getInstance("AES");

        // FIXME:: We should use AES256, if we ahve stong encrypt enabled.

        keyGen.init(128); // 128 bit legal export restrictions EMEA,APJ
        SecretKey key = keyGen.generateKey();
        byte[] raw = key.getEncoded();

        Base64.encodeToFile(raw, file.getAbsolutePath());
        return raw;
    }


    byte[] genRandom(int num) {
        byte[] data = new byte[num];
        SecureRandom random = new SecureRandom();
        random.nextBytes(data);
        return data;
    }


    public String encryptString(String input)  {
        try { 
            //byte[] encData = encryptBin(input.getBytes());
            byte[] encData = encryptWithSalt(input.getBytes());
            return Base64.encodeBytes(encData);
        }
        catch (InvalidKeyException e) {
            throw new RuntimeException(e);
        }
        catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    public String decryptString(String input)  {
        
        try { 
            byte[] encData = Base64.decode(input);

            //byte[] rawData = decryptBin(encData);
            byte[] rawData = decryptWithSalt(encData);
            if (rawData==null) return null;
            String outputStr = new String(rawData);

            return outputStr;

        }
        catch (InvalidKeyException e) {
            throw new RuntimeException(e);
        }
        catch (IOException e) {
            throw new RuntimeException(e);
        }

    }



    int saltLen = 4;

    public byte[] encryptWithSalt(byte[] input) throws InvalidKeyException, IOException {	   

        byte[] newData = new byte[input.length + saltLen];

        byte[] salt = genRandom(saltLen);
        System.arraycopy(salt, 0, newData, 0, salt.length);

        System.arraycopy(input, 0, newData, 4, input.length);

        return encryptBin(newData);


    }

    public byte[] decryptWithSalt(byte[] input) throws InvalidKeyException, IOException {
        byte[] rawData = decryptBin(input);

        int origLen = rawData.length - saltLen;
        if (origLen<0) return null;
        byte[] origData = new byte[origLen];

        System.arraycopy(rawData, 4, origData, 0, origData.length);

        return origData;
    }


    public byte[] encryptBin(byte[] rawData) throws InvalidKeyException, IOException  {

        cipher.init(Cipher.ENCRYPT_MODE, secretKeySpec);

        ByteArrayInputStream bIn = new ByteArrayInputStream(rawData);

        CipherInputStream cIn = new CipherInputStream(bIn, cipher);
        ByteArrayOutputStream bOut = new ByteArrayOutputStream();

        // TODO: Bigger chunks would be better.
        int ch;
        while ((ch = cIn.read()) >= 0) {
            bOut.write(ch);
        }

        byte[] encData = bOut.toByteArray();

        return encData;
    }

    public byte[] decryptBin(byte[] encData) throws InvalidKeyException, IOException  {
        cipher.init(Cipher.DECRYPT_MODE, secretKeySpec);

        ByteArrayOutputStream bOut = new ByteArrayOutputStream();
        CipherOutputStream cOut = new CipherOutputStream(bOut, cipher);
        cOut.write(encData);
        cOut.close();

        byte[] rawData = bOut.toByteArray();

        return rawData;

    }	   

    public static void main(String[] args) throws Exception {
        DataEncryptor se = new DataEncryptor(); 
        String plainText = "Example message";
        String encrypted = se.encryptString(plainText);
        String backAgain = se.decryptString(encrypted);
        System.out.println(backAgain);
    }

}

