package com.ca.smdeletesession;


import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.io.Serializable;
import java.text.DecimalFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Properties;
import java.util.TimeZone;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.GZIPOutputStream;

import javax.naming.Context;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;
import javax.naming.directory.SearchControls;
import javax.naming.ldap.InitialLdapContext;
import javax.naming.ldap.LdapContext;

import com.sf.smutils.util.DataEncryptor;
import com.sf.smutils.util.RuntimeProperties;
import com.unboundid.asn1.ASN1OctetString;
import com.unboundid.ldap.sdk.AddRequest;
import com.unboundid.ldap.sdk.Attribute;
import com.unboundid.ldap.sdk.BindResult;
import com.unboundid.ldap.sdk.DeleteRequest;
import com.unboundid.ldap.sdk.DereferencePolicy;
import com.unboundid.ldap.sdk.Filter;
import com.unboundid.ldap.sdk.LDAPConnection;
import com.unboundid.ldap.sdk.LDAPConnectionOptions;
import com.unboundid.ldap.sdk.LDAPException;
import com.unboundid.ldap.sdk.LDAPResult;
import com.unboundid.ldap.sdk.ResultCode;
import com.unboundid.ldap.sdk.SearchRequest;
import com.unboundid.ldap.sdk.SearchResult;
import com.unboundid.ldap.sdk.SearchResultEntry;
import com.unboundid.ldap.sdk.SearchScope;
import com.unboundid.ldap.sdk.controls.SimplePagedResultsControl;
import com.unboundid.ldap.sdk.controls.SubtreeDeleteRequestControl;
import com.unboundid.util.ssl.SSLUtil;
import com.unboundid.util.ssl.TrustAllTrustManager;

public class SessionStore_LdapDelete2 {

	//public class Main {
	public static void main(String[] args) {
		
		try {
			
		    SessionStore_LdapDelete2 main = new SessionStore_LdapDelete2();
			main.runMain();
			
		}
		catch (Exception e) {
			e.printStackTrace();
		}
			
	}

    boolean debugTrace = false;
    boolean debugTraceAll = false;
    boolean ldapTestMode = false;
    boolean ldapUsePageControl = true; 
	
    
    LDAPConnection ldapConnection = null;
    
    
    public void runStats() {
        init(); 
        
        MonitorThread monitorThread = new MonitorThread();
        monitorThread.setName("Stats");
        monitorThread.printStatus();
        
    }
    void runMain() {

        init(); 
        
        // Check if the prior instance is still running, and if so just exit.
        if (isStillRunning()) {
            return;
        }

        // We are continuing so modify the timestamp file and continue.
        touchFileTimeStamp();               
        
        printConfig(); 
        
        connect();

        runDeleteCount();
        
        if (isRunDelete()) {
            
            startMonitorThread();


            // If we are using page control - it will loop through all the entries found
            do {
                runDeleteSearch();
                waitForChildrenToComplete();
                reportSearchTotals();
            }
            while (!isRunCompleted());

            stopMonitorThread();
            
        }
        reportGrandTotals();
        
        close();
        
        logMessage("============== Ending SMDeleteSession ==================");
        
        closeAndRollLogFiles();
        
        removeKeepAliveFile();
    }

    /**
     * Not really used 
     */
    boolean stoppingChildren = false;
    boolean isStoppingChildren() { return stoppingChildren; } 

    void waitForChildrenToComplete() {
        stoppingChildren = true;
        for (DeleteWorkerThread workerThread : workerList) {
            while (workerThread.isRunning()) {
                try {
                    Thread.sleep(200);
                }
                catch (InterruptedException e) {}
            }
        }        
    }


    /**
     * Increment counters, on slight change of duplicate update by different threads, use synchronsed on variable. 
     */
    String countTag = "CounterLock";
    
    void incTotalCountAndTransTime(long blockDeletedCount, long blockTransTime) {
        
        synchronized (countTag) {
            totalTransTime += blockTransTime;
            totalDeletedCount += blockDeletedCount;
        }
    }

    void reportSearchTotals() {
        if (debugTrace) {

            long avTotalTransTime = (totalDeletedCount == 0) ? 0 : (totalTransTime / totalDeletedCount); 
            
            logMessage("");
            logMessage("Total Entries read " + formatNumber(totalEntriesRead) + " Entries deleted " + formatNumber(totalDeletedCount) + " av (" + avTotalTransTime + "ms)" );
        }

    }
    
    void reportGrandTotals() {
        if (debugTrace) {

            //long avTotalTransTime = (totalDeletedCount == 0) ? 0 : (totalTransTime / totalDeletedCount); 
            
            //logMessage("");
            //logMessage("Total Entries read " + totalEntriesRead + " Entries deleted " + totalDeletedCount + " av (" + avTotalTransTime + "ms)" );
            logMessage("============== Ending  Search  ====================");
        }

    }
    /**
     * Conditions under which we check if we have completed deleting all the entries. 
     */
    
    boolean isRunCompleted() {
        
        // TODO Check on this - there is a choice - we can exit, which is nice to reset everything. 
        //   or we can wait until totalDeletedCount == 0 which means process will be alive longer. 
        //   currently we are exiting for stability reasons.
        
        // If we are using page control, we will have done one loop through all entries - and exit then.
        
        if (!repeatRunUntilExpiredQueueCleared) {
            if (ldapUsePageControl) {
                return true;
            }
        }
        
        // If we are not using page controls (bug in horizontal partition CADir) then we keep reading the first page 
        // records until we finally do not find any entries to delete.
        
        reRunDeleteCount();
        
        if (!isRunDelete()) {
            return true; 
        }
        
        //if (totalDeletedCount == 0) {
        //    return true;
        //}
        
        // If we are here, then we are not using pages controls, and the last read of page entries, found some to delete
        // so we loop round again. 
        return false;
    }
    
    long keepAliveCheckTime = 120;
    String keepAliveFileName  = ".keepAliveFile";

    String traceLogFileName = null;
    String logFileName = null;
    
    boolean logToStdout = true;
    boolean traceLogToStdout = false;
    
    ArrayList<String> ldapMonitorSearchList = null;
    String ldapCountAttribute = null;
    long monitorRefreshTime = 1000; 
    
    long runDeleteThreshold = 1000; 

    boolean syncDeleteCountersAtEndOfBlock = false; 

    DataEncryptor encryptor = new DataEncryptor();
    
    void init() {
        getPolicyStoreConnectProperties();
        
        totalEntriesRead = 0;
        totalDeletedCount = 0;
        totalTransTime = 0;

        
        debugTrace = connectProperties.getPropertyAsBoolean("debugTrace", true);
        debugTraceAll = connectProperties.getPropertyAsBoolean("debugTraceAll", false);
        ldapTestMode = connectProperties.getPropertyAsBoolean("ldapTestMode", false);
        ldapUsePageControl = connectProperties.getPropertyAsBoolean("ldapUsePageControl", true);
        pageSize = connectProperties.getPropertyAsInt("pageSize", 1000);

        keepAliveCheckTime = connectProperties.getPropertyAsLong("keepAliveCheckTime", 60 );
        keepAliveFileName  = connectProperties.getProperty("keepAliveFileName", ".keepAliveFile");
        
        deleteWorkerWaitTimeout = connectProperties.getPropertyAsLong("deleteWorkerWaitTimeout", 120);
        deleteWorkerMaxThreads = connectProperties.getPropertyAsInt("deleteWorkerMaxThreads", 20);
        runDeleteThreshold = connectProperties.getPropertyAsLong("runDeleteThreshold", 1000);
        deleteRequestBlockSize = connectProperties.getPropertyAsInt("deleteRequestBlockSize", pageSize); 
        
        
        ldapDeleteRoot = connectProperties.getProperty("ldapDeleteRoot");
        ldapDeleteSearch = connectProperties.getProperty("ldapDeleteSearch");

        sizeLimit = connectProperties.getPropertyAsInt("sizeLimit", 0); 
        

        providerUrl = connectProperties.getProperty("ldapProviderUrl");
        userDn = connectProperties.getProperty("ldapUserDn", null);
        
        // Try an encrypted password first
        password = connectProperties.getProperty("ldapUserPasswordEnc", null);        
        if (password != null) {
            try { 
                password = encryptor.decryptString(password);
            }
             catch (Exception e) {
                 System.out.println("Unable to decrypt password from proeprties file");
             }
        }
        
        if (password == null) {
            password = connectProperties.getProperty("ldapUserPassword", null);
        }
        followReferrals = connectProperties.getPropertyAsBoolean("followReferrals", false);

        timeLimit = connectProperties.getPropertyAsInt("timeLimit", 30);        

        traceLogFileName = connectProperties.getProperty("traceLogFile", "log/smdeletesession_trace.log");
        logFileName = connectProperties.getProperty("logFile", "log/smdeletesession.log");

        logToStdout = connectProperties.getPropertyAsBoolean("logToStdout", true);
        traceLogToStdout = connectProperties.getPropertyAsBoolean("traceLogToStdout", false);
        
                
        ldapMonitorSearchList = connectProperties.getPropertiesList("ldapMonitorSearch");
        ldapCountAttribute = connectProperties.getProperty("ldapMonitorCountAttribute", "dxEntrycount");
        monitorRefreshTime = connectProperties.getPropertyAsLong("ldapMonitorRefreshTime", 120);
        monitorRefreshTime *= 1000;

        // Internal settings, not documented or printed currently 
        
        repeatRunUntilExpiredQueueCleared = connectProperties.getPropertyAsBoolean("repeatRunUntilExpiredQueueCleared", true);

        syncDeleteCountersAtEndOfBlock = connectProperties.getPropertyAsBoolean("syncDeleteCountersAtEndOfBlock", false);

    }
    
    
    String versionStr = "SMDeleteSession-V1.0.3";
    void printConfig() {
        if (debugTrace) { 
            logMessage("============== Starting SMDeleteSession ==================");
            logMessage("");
            logMessage("============== Delete Search Settings ====================");
            logMessage("Version : " + versionStr);
            logMessage("");
            logMessage("Delete   Root: " + ldapDeleteRoot );
            logMessage("Delete Search: " + ldapDeleteSearch);
            logMessage("");
            logMessage("");
            logMessage("Test Mode (does not delete) : " + ldapTestMode);
            logMessage("ldapUsePageControl          : " + ldapUsePageControl);
            logMessage("  pageSize: " + pageSize);
            logMessage("  timeLimit:" + timeLimit + "sec ");
            logMessage("  sizeLimit:" + sizeLimit);
            logMessage("");
            logMessage("keepAliveCheckTime          : " + keepAliveCheckTime);
            logMessage("keepAliveFileName           : " + keepAliveFileName);
            logMessage("");            
            logMessage("runDeleteThreshold          : " + runDeleteThreshold); 
            logMessage("deleteWorkerWaitTimeout     : " + deleteWorkerWaitTimeout);
            logMessage("deleteWorkerMaxThreads      : " + deleteWorkerMaxThreads);
            logMessage("deleteRequestBlockSize      : " + deleteRequestBlockSize);
            logMessage("");
            logMessage("ldapProviderUrl             : " + providerUrl);
            logMessage("ldapUserDn                  : " + userDn);
            logMessage("ldapUserPassword            : " + "<not shown>");
            logMessage("followReferrals             : " + followReferrals);

            logMessage("");
            
            logMessage("traceLogFileName            : " + traceLogFileName);
            logMessage("logFileName                 : " + logFileName);

            logMessage("logToStdout                 : " + logToStdout);
            logMessage("traceLogToStdout            : " + traceLogToStdout);
            logMessage("");

            for (int i =0 ; i < ldapMonitorSearchList.size(); i++) {
                logMessage("ldapMonitorSearch_" + i + " : "  + ldapMonitorSearchList.get(i));
            }
            logMessage("");
                    
            logMessage("ldapCountAttribute          : " + ldapCountAttribute) ;
            logMessage("monitorRefreshTime          : " + (monitorRefreshTime / 1000) + " sec" );

            logMessage("");
            
            logMessage("repeatRunUntilExpiredQueueCleared : " + repeatRunUntilExpiredQueueCleared) ;
            logMessage("syncDeleteCountersAtEndOfBlock    : " + syncDeleteCountersAtEndOfBlock) ;

            syncDeleteCountersAtEndOfBlock = connectProperties.getPropertyAsBoolean("syncDeleteCountersAtEndOfBlock", false);
            
        }

    }
    
    File keepAliveFile = null;
    
    File getKeepAliveFile() {
        
        if (keepAliveFile == null) {
            keepAliveFile = new File(keepAliveFileName);
        }
        return keepAliveFile;
    }

    void removeKeepAliveFile() { 
        File file = getKeepAliveFile();

        if (file.exists()) {
            try { 
                file.delete();
            }
            catch (Throwable e) {}
        }
    }

    
    boolean isStillRunning() {

        File file = getKeepAliveFile();
        
        if (!file.exists()) {
             return false;
        }

        
        long lastModified = file.lastModified();
        long keepAliveTime = keepAliveCheckTime * 1000; 

        long timeDiff = System.currentTimeMillis() - lastModified;
        
        if (timeDiff > keepAliveTime) { 
            return false;
        }
        
        if (debugTrace) {
            logMessage("Previous run is still going or completed recently (modified " + formatNumber(timeDiff) + " ms ok - within " + formatNumber(keepAliveTime)  + " ms) - exiting");
        }
        
        return true;
    }
    
    /**
     * Update timestamp on the file.
     */
    
    void touchFileTimeStamp() {
        File file = getKeepAliveFile();

        try { 
            FileWriter fw = new FileWriter(file);
            Date date = new Date(System.currentTimeMillis());
            
            fw.append(date.toString());
            fw.close();
        }
        catch (IOException e) {}
        
    }

    void connect() {
        ldapConnection = newConnection();
    }

    String providerUrl = null;
    String userDn = null;
    String password = null;
    boolean followReferrals = false;
    
    LDAPConnection newConnection() {

        try { 

            Pattern urlPattern = Pattern.compile("(\\w+)://(\\S+):(\\w+)");
            
            Matcher matcher = urlPattern.matcher(providerUrl);
            
            if (!matcher.find()) {
                throw new RuntimeException("Unable to parse ldap[s]://<hostaddress>:<port> from :" + providerUrl);
            }
            
            String protocol = matcher.group(1);
            String address = matcher.group(2);
            int port = 389;
            
            try { 
                String strPort = matcher.group(3);
                port = Integer.parseInt(strPort);
                
            }
            catch (NumberFormatException e) {
                throw new RuntimeException("Unable to parse port from " + providerUrl, e);
            }
            
            boolean isSSL = protocol.toLowerCase().equals("ldaps");
            
            LDAPConnection ldapConnection = null;
            
            if (isSSL) {
                 SSLUtil sslUtil = new SSLUtil(new TrustAllTrustManager());
                ldapConnection =    new LDAPConnection(sslUtil.createSSLSocketFactory());
            }
            else {
                ldapConnection = new LDAPConnection();
            }

            ldapConnection.connect(address, port);
            BindResult bindResult = ldapConnection.bind(userDn, password);

            LDAPConnectionOptions options = new LDAPConnectionOptions();
            
            options.setFollowReferrals(followReferrals);
            
            ldapConnection.setConnectionOptions(options);
            
            String controlOID = SimplePagedResultsControl.PAGED_RESULTS_OID;
            //  if(!SupportedFeature.isControlSupported(ldapConnection,controlOID)) {
            //    return ResultCode.UNWILLING_TO_PERFORM;
            //  }
            
            return ldapConnection;

        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }
	
	RuntimeProperties connectProperties = null;
	
    public RuntimeProperties getPolicyStoreConnectProperties() {
        if (connectProperties == null) {
            
            Properties properties = new Properties();

            File file = new File("policyStoreConnect.properties");
            if (file.exists()) {
                try {
                    properties.load(new FileInputStream(file));
                } catch (FileNotFoundException e) {
                } catch (IOException e) {
                }
            }
            connectProperties = new RuntimeProperties();
            connectProperties.setProperties(properties);
        }
        
        return connectProperties;
    }

    
    int sizeLimit = 0;
	int timeLimit = 30;
	
	long transTime = 0;
	
	long totalTransTime = 0;

    int totalDeletedCount = 0; 
	
	
    SearchRequest buildSearchRequest(String baseDN, String search) throws LDAPException {
		SearchScope scope = SearchScope.SUB; //SearchScope.SUBORDINATE_SUBTREE;
	      Filter filter = Filter.create(search);
	      
	      SearchRequest searchRequest;
          searchRequest = new SearchRequest(baseDN,scope,filter,SearchRequest.NO_ATTRIBUTES); 
          //String[] requestedAttributes = new String[0];
	      //searchRequest = new SearchRequest(baseDN,scope,filter,requestedAttributes);
	      searchRequest.setSizeLimit(sizeLimit);
	      searchRequest.setTimeLimitSeconds(timeLimit);
	      
	      return searchRequest;
    }

    SearchRequest buildCountSearchRequest(String baseDN, String search) throws LDAPException {
        SearchScope scope = SearchScope.SUB; //SearchScope.SUBORDINATE_SUBTREE;
          Filter filter = Filter.create(search);
          
          SearchRequest searchRequest;
          //searchRequest = new SearchRequest(baseDN,scope,filter,SearchRequest.NO_ATTRIBUTES); 

          String[] requestedAttributes = new String[1];
          requestedAttributes[0] = ldapCountAttribute;
          
          searchRequest = new SearchRequest(baseDN,scope,filter,requestedAttributes);
          searchRequest.setSizeLimit(sizeLimit);
          searchRequest.setTimeLimitSeconds(timeLimit);
          
          return searchRequest;
    }
    
    long getEntryCount(LDAPConnection ldapConnection, String ldapSearch) {
        
        try {
            ldapSearch = substituteLdapTimeStamp(ldapSearch);
            SearchRequest searchRequest = buildCountSearchRequest(ldapDeleteRoot, ldapSearch);
            SearchResult searchResult = ldapConnection.search(searchRequest);

            long result = 0;

            for (SearchResultEntry searchEntry : searchResult.getSearchEntries()) {
                long entryValue = searchEntry.getAttributeValueAsLong(ldapCountAttribute);
                result += entryValue;
            }

            return result;
        }
        catch (LDAPException e) {
            /*
            LDAPResult deleteResult = deleteResult = le.toLDAPResult();
            ResultCode resultCode = le.getResultCode();
            String errorMessageFromServer = le.getDiagnosticMessage();
            
            logMessage("Error : " + resultCode + ": " +  errorMessageFromServer);
             */
            throw new RuntimeException(e);
        }
    }
    
    
    String ldapDeleteRoot = null;
    String ldapDeleteSearch = null;
    int pageSize = -1; 
    int totalEntriesRead = 0;

    String substituteLdapTimeStamp(String ldapSearch) {
        // If we find %TIME then substitute the current LDAP GMT time 
        
        if (ldapSearch.indexOf("%TIME") == -1) {
            return ldapSearch;
        }
        
        ldapSearch = ldapSearch.replace("%TIME", getCurrentTimeAsLdap());
        
        return ldapSearch;
   }
    
    long origEntryCount = -1; 
    long lastOrigEntryCount = -1;
    String ldapOriginalSearch = null;
    
    boolean runDelete = false; 
    boolean isRunDelete() { return runDelete;  } 
    void setRunDelete(boolean runDelete) { this.runDelete = runDelete;  } 
    
    void runDeleteCount() {

        // Count the number of entries we need to delete - we store the Orig search as well. 
        ldapOriginalSearch = ldapDeleteSearch;
        ldapOriginalSearch = substituteLdapTimeStamp(ldapOriginalSearch);
        
        origEntryCount = getEntryCount(ldapConnection, ldapOriginalSearch);
        lastOrigEntryCount = origEntryCount;
        
        // Check if we are over the threashold value to run the delete process.  
        boolean runDelete = (origEntryCount > runDeleteThreshold );
        setRunDelete(runDelete);

        // Print out the status to the logs. 
        
        String label = "ExpiredSessionCount";
        String msg = label + " " + ldapOriginalSearch + " : " + formatNumber(origEntryCount) + " entries";

        logMessage("");
        logMessage(msg);
        
        if (isRunDelete()) {
            msg = "Will run SMDeleteSession process - (found " + origEntryCount + " expired smSessions > "  + runDeleteThreshold + ")"; 
        }
        else {
            msg = "Will NOT run SMDeleteSession process - (only found " + origEntryCount + " expired smSessions <= "  + runDeleteThreshold + ")"; 
            
        }
        logMessage("");
        logMessage(msg);
        
    }

    /**
     * Here we rerun the orignal delete count - with the idea of trying to workout if non-paged mode should keep 
     * going of if it is complete now. 
     */
    
    boolean repeatRunUntilExpiredQueueCleared = false; 
    
    void reRunDeleteCount() {

        // Count the number of entries we need to delete - from the original search.
        
        // For non paged search we have option to rerun the timestamp substituion, rather than keep using the 
        // original one. 
        if (repeatRunUntilExpiredQueueCleared ) { 
            ldapOriginalSearch = ldapDeleteSearch;
            ldapOriginalSearch = substituteLdapTimeStamp(ldapOriginalSearch);
        }
        
        // Count the number of entries we need to delete - we store the Orig search as well.
        lastOrigEntryCount = getEntryCount(ldapConnection, ldapOriginalSearch);
        
        // Check if we are over the threashold value to run the delete process.  
        boolean runDelete = (lastOrigEntryCount > runDeleteThreshold );
        setRunDelete(runDelete);

        // Print out the status to the logs. 

        String msg = null;
        //String label = "RemainExpiredSessionCount";
        //String msg = label + " " + ldapOriginalSearch + " : " + formatNumber(lastOrigEntryCount) + " entries";

        //logMessage("");
        //logMessage(msg);
        
        if (isRunDelete()) {
            msg = "Will re-run SMDeleteSession process - (still " + lastOrigEntryCount + " expired smSessions)"; 
        }
        else {
            msg = "Will NOT re-run SMDeleteSession process - (only " + lastOrigEntryCount + " expired smSessions)"; 
            
        }
        logMessage("");
        logMessage(msg);
        
    }
    
    boolean firstRun = true; 
    void runDeleteSearch() {
    
    	try { 
            
    		SearchRequest searchRequest = buildSearchRequest(ldapDeleteRoot, ldapOriginalSearch);

            if (debugTrace) {
                logMessage("============== Starting Search  ====================");
                if (firstRun) { 
                    logMessage("");
                    logMessage("Delete   Root: " + ldapDeleteRoot );
                    logMessage("Delete Search: " + ldapDeleteSearch);
                    logMessage("");
                    
                    firstRun = false;
                }
                //logMessage("pageSize: " + pageSize + " timeLimit:" + timeLimit + " sec");
            }
    		
    		ASN1OctetString cookie = null;

			
    		do {
    		    
    		    touchFileTimeStamp();    		    
    			/*
    			 * Set the simple paged results control (if the cookie is null
    			 * this indicates the first time through the loop).
    			 */
    			final SimplePagedResultsControl simplePagedResultsRequestControl = new SimplePagedResultsControl(pageSize,cookie);
    			
    			if (ldapUsePageControl) { // Optional to use page control
    			    searchRequest.setControls(simplePagedResultsRequestControl);
    			}

    			/*
    			 * Issue the search request:
    			 */
    			SearchResult searchResult = null;
    			
    			try {
    			    searchResult = ldapConnection.search(searchRequest);
    			    
    			}
    	        catch (LDAPException e) {
    	            LDAPResult failResult = e.toLDAPResult();
    	            if (!(failResult instanceof SearchResult)) {
                        throw new RuntimeException(e);
    	            }
    	            
    	            searchResult = (SearchResult) failResult;

    	            ResultCode resultCode = e.getResultCode();
    	            String errorMessageFromServer = e.getDiagnosticMessage();
    	            if (errorMessageFromServer == null) errorMessageFromServer = "";

    	            String errLevel = "Error";
    	            if (resultCode == ResultCode.ADMIN_LIMIT_EXCEEDED || 
    	                    resultCode == ResultCode.SIZE_LIMIT_EXCEEDED ||
    	                    resultCode == ResultCode.TIME_LIMIT_EXCEEDED ) {
    	                errLevel = "Warning";
    	            }

    	            
    	            logMessage(errLevel + " : " + resultCode + ": " +  errorMessageFromServer + " entries : " + formatNumber(searchResult.getEntryCount()));
    	        }

    			if (debugTraceAll) {
                    String msg = String.format("searchRequest transmitted, " + "entries returned: %d", Integer.valueOf(searchResult.getEntryCount()));
    			    if (ldapUsePageControl) {
    			        msg += ", page size : " +  pageSize; 
    			    }
   			        logTraceMessage("");
                    logTraceMessage(msg);
                }

                totalEntriesRead += searchResult.getEntryCount();
    			
    			deleteEntries(searchResult);
                

				
    			// Get the cookie from the paged results control.
    			cookie = null;
    			if (ldapUsePageControl) {
    			    final SimplePagedResultsControl c = SimplePagedResultsControl.get(searchResult);
                    if(c != null) {
                        cookie = c.getCookie();
                    }
    			}

    			// Without page control cookie will always be null 
    		} while((cookie != null) && (cookie.getValueLength() > 0));
			
    	}
    	catch (Exception e) {
    		throw new RuntimeException(e);
    	}
    }

    SimpleDateFormat sdf = null; 

    SimpleDateFormat getLDAPDateFormat() {
        if (sdf == null) {
            sdf = new SimpleDateFormat("yyyyMMddHHmmss");
            sdf.setTimeZone(TimeZone.getTimeZone("GMT"));
        }
        return sdf;
        
    }
    public String getCurrentTimeAsLdap() {
        Date date = new Date();
        
        String dateStr = getLDAPDateFormat().format(date);
        dateStr +="Z";
        
        return dateStr;
    }
    
    public Date parseLdapDate(String ldapDate){

        try {
            return getLDAPDateFormat().parse(ldapDate);
        } catch (ParseException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        return null;
    }
    
    
    long deleteRequestBlockSize = 1000;
    
    void deleteEntries(SearchResult searchResult ) {

        // Add all dn's found to an array. 
        ArrayList<String> deleteList = new ArrayList<String>(); 
        for (SearchResultEntry searchEntry : searchResult.getSearchEntries()) {
            String dn = searchEntry.getDN();
            deleteList.add(dn);
            
            // For every group we get of 1000 records we spin (or wait) for free bg thread to process it
            // mainly this is for non-paged results where we read maybe 10,000 entries, rather than 1,000 entries in a page.
            if (deleteList.size() >= deleteRequestBlockSize) {
                // Run background thread to delete the entries. 
                runDelete(deleteList);
                deleteList = new ArrayList<String>(); 
            }
        }
        
        runDelete(deleteList);
        
    }

    void runDelete(ArrayList<String> deleteList ) {
        // If we dont find any just return.
        if (deleteList.size() == 0) {
            return;
        }
        
        // Run background thread to delete the entries. 
        DeleteWorkerThread worker = getWorker();
        worker.runDelete(deleteList);
        
    }
    void XdeleteEntries(SearchResult searchResult ) {

        // Add all dn's found to an array. 
        ArrayList<String> deleteList = new ArrayList<String>(); 
        for (SearchResultEntry searchEntry : searchResult.getSearchEntries()) {
            String dn = searchEntry.getDN();
            deleteList.add(dn);
        }
        
        // If we dont find any just return.
        if (deleteList.size() == 0) {
            return;
        }
        
        // Run background thread to delete the entries. 
        DeleteWorkerThread worker = getWorker();
        worker.runDelete(deleteList);
        
    }
    
    ArrayList<DeleteWorkerThread> workerList = new ArrayList<DeleteWorkerThread>(); 
    LinkedBlockingQueue<DeleteWorkerThread> workerQueue = new LinkedBlockingQueue<DeleteWorkerThread>(); 

    long nextTimeToAddWorker = -1;
    long deleteWorkerWaitTimeout = 120;
    int deleteWorkerMaxThreads = 20;

    void checkNeedToAddWorkers() {
        while (isNeedToAddWorker()) {
            newWorker();
        }
        
    }
    
    boolean isNeedToAddWorker() {

       
        // Check if we have an available worker already - then do not need to add one.
        if (workerQueue.size() > 0) { 
            return false;
        }

        // Check if we have exceeded our worker thread limit.
        if (workerList.size() >= deleteWorkerMaxThreads) {
            return false; 
        }
        
     // This is first call - we are at zero - return true, and set next time to add thread to 2min in the future.
        if (nextTimeToAddWorker == -1 ) {  
            nextTimeToAddWorker = System.currentTimeMillis() + deleteWorkerWaitTimeout * 1000;
            return true; 
        }
        
        // This is incremental call - if 2min expired, then inc next expire time by 2min and return true.
        // Allows for multiple starts if time is 4min behind.
        long time = System.currentTimeMillis() - nextTimeToAddWorker;
        if (time > 0) {
            nextTimeToAddWorker += deleteWorkerWaitTimeout * 1000;
            return true;
        }
        
        // Otherwise we do not need another worker.
        return false;
    }
    
    DeleteWorkerThread getWorker() {

        DeleteWorkerThread thread = null;


        // Determine how long to wait for queue response, we currently wait 1/2 the time needed to add a worker thread. 
        long checkTime = deleteWorkerWaitTimeout / 2;
        if (checkTime == 0) checkTime = 1;

        long startTime = System.currentTimeMillis();
        while (thread == null) {

            checkNeedToAddWorkers();
            try { 
                thread = workerQueue.poll(deleteWorkerWaitTimeout, TimeUnit.SECONDS);
            }
            catch (InterruptedException e) {}

            if (debugTraceAll) {
                if (thread == null) {
                    long waitTime = System.currentTimeMillis() - startTime;  
                    logTraceMessage(" Timeout waiting for worker : " + waitTime + "ms");
                }
            }

        }

        return thread;
    }
    
    void newWorker() {
        
        String name = "deleteWorker-" + (workerList.size() + 1);

        DeleteWorkerThread thread = new DeleteWorkerThread();
        thread.setName(name);
        
        workerList.add(thread);
        workerQueue.add(thread);
        
        logMessage("Adding a new deleteWorker thread : " + name + " [ now have " + workerList.size() + " deleteWorker threads running ]");
        
    }
    
    void returnWorker(DeleteWorkerThread worker) {
        workerQueue.add(worker);
    }

    void close() {
        
        // Close all the worker threads ldap connections.
        for (DeleteWorkerThread thread: workerList) {
            thread.close();
        }
        
        // Close the search connection.
        ldapConnection.close();
    }
    
    void closeAndRollLogFiles() {
        
        // Close the log files ... 
        if (logWriter != null) {
            logWriter.flush();
            logWriter.close();
            
        }
        
        if (traceLogWriter != null) {
            traceLogWriter.flush();
            traceLogWriter.close();
        }
    }

    MonitorThread monitorThread = null; 
    
    void startMonitorThread() {
        
        if (monitorThread == null) {
            String name = "MonitorThread";

            if (debugTrace) {
                logMessage("Starting Monitor thread : " + name);
            }
            
            monitorThread = new MonitorThread();
            monitorThread.setName(name);
            monitorThread.start();
        }
        
    }
    
    void stopMonitorThread() {
        if (monitorThread != null ) {
            monitorThread.interrupt();
            monitorThread = null;
        }
        
        /* Alternative way 
        // Wait for monitor to stop.
        try {
            monitorThread.join();
        }
        catch (InterruptedException e) {}
        */


    }
    
    
	private static SearchControls getSimpleSearchControls() {
	    SearchControls searchControls = new SearchControls();
	    searchControls.setSearchScope(SearchControls.SUBTREE_SCOPE);
	    searchControls.setCountLimit(0);
	    searchControls.setTimeLimit(30000);
	    //String[] attrIDs = {"objectGUID"};
	    //searchControls.setReturningAttributes(attrIDs);
	    return searchControls;
	}

    PrintWriter createPrintWriter(String fileName) {
        try { 
            
            String dateStr = getCurrentDateStr();
            
            String fullFileName = fileName + "." + dateStr;
            
            File file = new File(fullFileName);

            // Create parent log file directory if it does not exist. 
            File parentDir = file.getParentFile();
            if (!parentDir.exists()) {
                parentDir.mkdirs();
            }
            
            FileWriter fw = new FileWriter(file, true);   // We append to date stamped file.
            PrintWriter pw  = new PrintWriter(fw, true);
            
            return pw;
        }
        catch (FileNotFoundException e) {
            throw new RuntimeException(e);
        }
        catch (IOException e) {
            throw new RuntimeException(e);
        }
        
    }

	PrintWriter traceLogWriter = null;
		
    void logTraceMessage(String msg) {

        if (!debugTraceAll) {
            return;
        }
        
        
        String rptMsg = getCurrentTimeStampStr();
        rptMsg += " : ";
        rptMsg += Thread.currentThread().getName();
        rptMsg += " : ";
        rptMsg += msg;
        
        writeTraceLogMessage(rptMsg);

        if (traceLogToStdout) {
            System.out.println(rptMsg);
        }
        
    }

    void writeTraceLogMessage(String rptMsg) {
        
        if (!debugTraceAll) {
            return;
        }

        if (traceLogWriter == null) {
            traceLogWriter = createPrintWriter(traceLogFileName);
        }
        
        synchronized (traceLogWriter) {
            traceLogWriter.println(rptMsg);
        }
        
    }
    
    
    PrintWriter logWriter = null;
    
    void logMessage(String msg) {

        if (logWriter == null) {
            logWriter = createPrintWriter(logFileName);
        }
        
        String rptMsg = getCurrentTimeStampStr();
        rptMsg += " : ";
        rptMsg += Thread.currentThread().getName();
        rptMsg += " : ";
        rptMsg += msg;
        
        synchronized (logWriter) {
            logWriter.println(rptMsg);            
        }
        if (logToStdout) {
            System.out.println(rptMsg);
        }
        
        writeTraceLogMessage(rptMsg);


    }


    SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd");
    
    String getCurrentDateStr() {
        return dateFormat.format(new Date());
    }
    
	SimpleDateFormat timeStampFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
	
	String getCurrentTimeStampStr() {
	    return timeStampFormat.format(new Date());
	}
	
	
	DecimalFormat decimalFormat = new DecimalFormat( "#,###,###,##0" );
	
	String formatNumber(long num) {
	    return decimalFormat.format(num);
	}
	
	

    class DeleteWorkerThread implements Runnable {

       
        String name = null;
        public void setName(String name) { this.name = name; }
        
        LDAPConnection ldapConnection = null;
        
        long blockTransTime = 0;
        int blockDeletedCount = 0; 

        boolean running = false; 
        boolean isRunning() { return running; }
        void setRunning(boolean running) { this.running = running; } 
        
        DeleteWorkerThread() {
            init();
        }
        
        void init() {
            ldapConnection  = newConnection();

        }

        void close() {
            ldapConnection.close();
        }

        ArrayList<String> deleteList = new ArrayList<String> ();

        void runDelete(ArrayList<String> newDeleteList) {
            
            blockTransTime = 0;             
            blockDeletedCount = 0;
            
            // We copy them over, just to be sure they  
            deleteList = new ArrayList<String>();
            deleteList.addAll(newDeleteList); 

            setRunning(true);
            
            thread = new Thread(this);
            thread.setDaemon(true);
            thread.setName(name);
            thread.start();
        }
        
        Thread thread = null;

        
        public void run() {

            try {
                
                deleteEntries();
                
                if (debugTraceAll) { 
                    logTraceMessage("");
                    
                    long avBlockTransTime = (blockDeletedCount == 0) ? 0 : (blockTransTime / blockDeletedCount); 
                    logTraceMessage("Delete Thread " + Thread.currentThread().getName() + " Entries " + deleteList.size() + " Entries deleted " + totalDeletedCount + " this page" + blockDeletedCount + " av (" + avBlockTransTime + "ms)");
                }
                
                if (syncDeleteCountersAtEndOfBlock) {
                    incTotalCountAndTransTime(blockDeletedCount, blockTransTime);
                }
                

            }
            catch (Exception e) {
                e.printStackTrace();
            }
            finally {
                setRunning(false);
                returnWorker(this);
            }
            
        }
        
        void deleteEntries() { 
            for (String dn : deleteList) {
                ldapDelete(dn);
            }
        }

        
        void ldapDelete(String dn) {


            long startTime = System.currentTimeMillis();

            // Test mode we only report the entry not delete it.
            if (ldapTestMode) {  
                try { 
                    Thread.sleep(1);
                    //Thread.sleep(4);
                }
                catch (InterruptedException e) {} 
            }
            else {
                try  {
                    DeleteRequest deleteRequest = new DeleteRequest(dn);
                    deleteRequest.addControl(new SubtreeDeleteRequestControl());
                    LDAPResult deleteResult = ldapConnection.delete(deleteRequest);
                }
                catch (LDAPException le)  {
                    // The delete operation failed.
                    LDAPResult deleteResult = deleteResult = le.toLDAPResult();
                    ResultCode resultCode = le.getResultCode();
                    String errorMessageFromServer = le.getDiagnosticMessage();

                    logMessage("Error : " + resultCode + ": " +  errorMessageFromServer);
                } 
            }   
            
            transTime = System.currentTimeMillis() - startTime;
            blockTransTime += transTime;
            blockDeletedCount++;
            
            // More accurate stats to append after each delete. 
            if (!syncDeleteCountersAtEndOfBlock) {
                incTotalCountAndTransTime(1, transTime);
            }

            if (debugTraceAll) {
                String msg = "deleted entry:" + dn + "(" + transTime + "ms";
                if (ldapTestMode) {
                    msg = "not " + msg + " (test mode)";
                }
                logTraceMessage(msg);

            }
            
        }    
        
        
    }

    
    public class MonitorThread extends Thread {

        LDAPConnection ldapConnection = null;
        
        
        MonitorThread() {
            init();
        }
        
        void init() {
            //setDaemon(true);
            ldapConnection  = newConnection();
        }

        void close() {
            ldapConnection.close();
        }

        public void run() {

            try {
                
                while (true) {
                    
                    printStatus();
                    
                    if (monitorThread == null) {
                        return;
                    }
                    try {
                        Thread.sleep(monitorRefreshTime);
                    }
                    catch (InterruptedException e) {}
                }
                
            }
            catch (Exception e) {}
            finally {
                close();
            }
            
        }
        
        
        void printStatus() {
        
            // For first call get count of entries we need to delete.
            
            for (String ldapSearch : ldapMonitorSearchList ) {
                printEntryCount("Count", ldapSearch);
            }

            // TODO - dont really need this but left in case want to add latter.
            // We also ahve the original search, so we can use that to see how many are left too 
            // - but left this off currently as disabled 
            //if (ldapOriginalSearch != null) {
            //    logMessage("");
            //    printEntryCount("Count", ldapOriginalSearch);
            //}

            
            if (origEntryCount != -1 ) {
                logMessage("");
                logMessage("  Original # of smSessions found in search : " + formatNumber(origEntryCount) );
                
                // If not using paged entries then this value will be different.
                if (origEntryCount != lastOrigEntryCount) {
                    logMessage("  Last # of smSessions found in search     : " + formatNumber(lastOrigEntryCount) );
                }
                
                logMessage("  # of smSessions read                     : " + formatNumber(totalEntriesRead) );                
                logMessage("  # of smSessions deleted                  : " + formatNumber(totalDeletedCount));
                //logMessage("  Estimated # of sessions remaining        : " + formatNumber(origEntryCount - totalDeletedCount) );
                //logMessage("  Estimated # of sessions remaining        : " + formatNumber(lastOrigEntryCount) );

                // Do a query to find out how many we actually have left, since this program + policy server will 
                // be doing deletes of expired sessions. 

                String ldapRemainSearchCount = ldapOriginalSearch;
                if (repeatRunUntilExpiredQueueCleared ) { 
                    ldapRemainSearchCount = substituteLdapTimeStamp(ldapDeleteSearch);
                }
                
                long remainingEntries = getEntryCount(ldapConnection, ldapRemainSearchCount);
                logMessage("  # of smSessions remaining                : " + formatNumber(remainingEntries) );

                logMessage("");
                logMessage("  # of delete Worker threads                       : " + workerList.size() );
            }
            

        }

        void printEntryCount(String label, String ldapSearch) {
        
            try {
                ldapSearch = substituteLdapTimeStamp(ldapSearch);
                SearchRequest searchRequest = buildCountSearchRequest(ldapDeleteRoot, ldapSearch);
                SearchResult searchResult = ldapConnection.search(searchRequest);

                //searchResult.getSearchEntry(arg0)

                long result = 0;
                long entryCount = searchResult.getEntryCount();

                boolean hasPartitions = entryCount > 1;
                
                long thisEntryCount = 0;
                for (SearchResultEntry searchEntry : searchResult.getSearchEntries()) {
                    long entryValue = searchEntry.getAttributeValueAsLong(ldapCountAttribute);
                    
                    if (debugTraceAll) {
                        if (hasPartitions) {
                            thisEntryCount++;
                            logTraceMessage("     " + label + " Partition (" + thisEntryCount + ")  : "  + ldapSearch + " : " + formatNumber(entryValue) + " entries");
                        }
                    }
                    result += entryValue;
                }

                String msg = label + " " + ldapSearch + " : " + formatNumber(result) + " entries";
                if (hasPartitions) {
                    msg += " over (" + entryCount +" partitions )";
                }
                logMessage(msg);
                
            }
            catch (LDAPException e) {
                /*
                LDAPResult deleteResult = deleteResult = le.toLDAPResult();
                ResultCode resultCode = le.getResultCode();
                String errorMessageFromServer = le.getDiagnosticMessage();
                
                logMessage("Error : " + resultCode + ": " +  errorMessageFromServer);
                 */
                throw new RuntimeException(e);
            }
        }

        
    }
    
}

